__all__ = [
    "DEFAULT_XTB_NUM_CPU_CORES_CONSTRAINT",
    "CSV_DECIMAL_CHAR",
    "CSV_SEP_CHAR",
    "NOK",
    "OK",
    "SUFFIX_MOL",
    "SUFFIX_SDF",
    "SUFFIX_SMILES",
    "XTB_EXEC_NAME",
    "XTB_EXEC_STDERR_FILE_NAME",
    "XTB_EXEC_STDOUT_FILE_NAME",
    "XTB_OPTIMISED_XYZ_FILE_NAME",
]

CSV_DECIMAL_CHAR = '.'
CSV_SEP_CHAR = ';'
DEFAULT_XTB_NUM_CPU_CORES_CONSTRAINT = 4

OK = 0
NOK = 1

SUFFIX_MOL = ['.mol']
SUFFIX_SDF = ['.sdf']
SUFFIX_SMILES = ['.smi', '.smiles']

XTB_EXEC_NAME = "xtb"
XTB_EXEC_STDERR_FILE_NAME = "xtb.stderr"
XTB_EXEC_STDOUT_FILE_NAME = "xtb.stdout"
XTB_OPTIMISED_XYZ_FILE_NAME = "xtbopt.xyz"